/*  Copyright 2006 Jonas Minnberg

    This file is part of OldPlay - a portable, multiformat musicplayer.

    OldPlay is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    OldPlay is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OldPlay; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "FileList.h"
#include <SDL.h>
#include "SDLScreen.h"
#include "MusicPlayer.h"
#include "Equalizer.h"
#include "PlayList.h"
#include "PlayerApp.h"
#include "util.h"

#include "Input.h"

#ifdef _WIN32
#include <direct.h>
#endif

int plugin_filter_cb(void *data, const char *name)
{
	PlayerApp *app = (PlayerApp *)data;

	if(app->music_player->CanHandle(name))
		return 1;

	if(is_ext(name, ".zip"))
		return 1;

	return 0;
}

PlayerApp::~PlayerApp()
{
	char dst[8];
	config["REPEAT"] = repeat ? (char *)"1" : (char *)"0";
	config["SHUFFLE"] = shuffle ? (char *)"1" : (char *)"0";
	sprintf(dst, "%d", music_player->GetVolume());
	config["VOLUME"] = dst;
	config.Save();
}

PlayerApp::PlayerApp(char *fname)
{

	InitSettings(fname);

	pane = new ListPane(0, 0, 0, 0);
	songInfo = new SongInfoPane(0, 0, 0, 0);
	lcd = new LedControl(screen_cfg["LEDBMP"], 0, 0);
	eq = new Equalizer(50, 50, lcd->Width(), lcd->Width()/3);
	music_player = new MusicPlayer();
	//Input *input = new Input();
	music_player->SetVolume(atoi(config["VOLUME"]));

	lcd->SetLed(LedControl::SONGNO, 1);
	lcd->SetLed(LedControl::MAXSONG, 1);
	lcd->SetLed(LedControl::VOLUME, music_player->GetVolume());
	lcd->SetLed(LedControl::REPEAT, repeat);
	lcd->SetLed(LedControl::SHUFFLE, shuffle);

	play_list = new PlayList(&plugin_filter_cb, this);
	char curdir[256];
	getcwd(curdir, sizeof(curdir));
	file_list = new FileList(curdir, &plugin_filter_cb, this);
	zip_list = NULL;

//	repeat = shuffle = 0;

	fprintf(stderr, "PlayerApp constructed\n");

}

#define ABS(x) ((x) < 0 ? (-x) : (x))

void PlayerApp::Layout()
{
	int fh = g_screen.GetPropFont()->height;

//	settings = *g_screen.GetSettings();

	int w = g_screen.Width() - hmargin*2;
	int h = g_screen.Height() - vmargin*2;

	lcd->Move(w - lcd->Width() + hmargin, fh+sep + vmargin);
	lcd->Clear(true);

	pane->Move(hmargin, fh+sep + vmargin);
	pane->Resize(w - lcd->Width() - sep, h - (fh+sep));
	pane->Clear(true);

	songInfo->Move(w - lcd->Width() + hmargin, lcd->Height() + (fh+sep) + vmargin + sep);
	songInfo->Resize(lcd->Width(), fh * 9);
	songInfo->Clear(true);
	songInfo->Render();

	int y = lcd->Height() + (fh+sep) + sep + sep + fh * 9  + vmargin;
	eq->Move(w - lcd->Width() + hmargin, y);
	eq->Clear(true);

	Font *font = g_screen.GetPropFont();

	int len = font_text(font, "v0.95/sasq", -1, -1);

	int f = font->flags;
	font->flags |= FFLG_TRANSP;
	font_setcolor(0x000000, settings.colors[Screen::COL_BEVEL_MID]);
	font_text(font, "v0.95/sasq", w-len+1+hmargin, h-font->height+vmargin+1);
	font_setcolor(0xFFFFFF, settings.colors[Screen::COL_BEVEL_MID]);
	font_text(font, "v0.95/sasq", w-len+hmargin, h-font->height+vmargin);
	font->flags = f;

}

int last_sec = -1;
int paused = 0;
bool oldplaying = false;
int go_to = 0;

void PlayerApp::Render()
{
	if(paused)
	{
		if(((paused-1) % 15) == 0)
			lcd->SetLed(LedControl::PLAYING, (paused-1) % 30 );
		paused++;
	}

	if(playing != oldplaying)
	{
		lcd->SetLed(LedControl::PLAYING, playing );	
		oldplaying = playing;
	}

	int sec = music_player->GetCSeconds() / 10;

	if(last_sec != sec)
	{
		lcd->SetLed(LedControl::LENGTH, sec);
		last_sec = sec;
	}

	pane->Render();
	songInfo->Render();
	lcd->Render();

#ifdef GP2X
	if(!music_player->GetHeavy())
#endif
	{
		if(eq)
			eq->Render();
		RenderEq();
	}

}


bool PlayerApp::Update()
{
	music_player->SetSongEnd(repeat && play_list->Size());

	// Check if Musicplayer stopped playing
	if(playing && !music_player->Playing())
	{
		playing = false;

		fprintf(stderr, "Songswitch %d %d\n", play_list->Current(), (int)go_to);

		// Check if we should pick a new song from playlist
		if(play_list->Size() && /*(play_list->Current() >= 0) &&*/ (repeat || go_to))
		{
			if(go_to)
			{
				if(go_to < 0)
					play_list->Prev(repeat);
				else
					play_list->Next(repeat);
			}
			else
				play_list->Next(repeat);


			fprintf(stderr, "%d\n", play_list->Current());

			if(play_list->Current() != -1)
				PlaySong(play_list, play_list->Current() );
		}

		go_to = 0;

	}

	

	//songInfo->SetFormat(mp->GetFormat());
	//printf("%s\n", mp->GetFormat());

	return true;
}


bool PlayerApp::HandleKey(int key)
{
	switch(key)
	{
	case Input::NEXT:
		playing = true;
		music_player->Stop();
		go_to = 1;
		break;
	case Input::PREV:
		playing = true;
		music_player->Stop();
		go_to = -1;
		break;
	case Input::FF:
		music_player->SetTune(music_player->GetTune() + 1);
		lcd->SetLed(LedControl::SONGNO, music_player->GetTune()+1);
		songInfo->SetLength(music_player->GetLength());
		break;
	case Input::REW:
		music_player->SetTune(music_player->GetTune() - 1);
		lcd->SetLed(LedControl::SONGNO, music_player->GetTune()+1);
		songInfo->SetLength(music_player->GetLength());
		break;
	case Input::VOLUP:
		music_player->SetVolume(music_player->GetVolume() + 1);
		lcd->SetLed(LedControl::VOLUME, music_player->GetVolume());
		break;
	case Input::VOLDOWN:
		music_player->SetVolume(music_player->GetVolume() - 1);
		lcd->SetLed(LedControl::VOLUME, music_player->GetVolume());
		break;
	case Input::QUIT:
		SDL_CloseAudio();
		SDL_Quit();
		return false;
	case Input::SHUFFLE:
		shuffle = shuffle ? false : true;
		lcd->SetLed(LedControl::SHUFFLE, shuffle);
		play_list->SetShuffle((bool)shuffle);
		break;
	case Input::REPEAT:
		repeat = repeat ? false : true;
		lcd->SetLed(LedControl::REPEAT, repeat);
		break;
	case Input::PAUSE:
		paused = music_player->Pause();
		break;
	case Input::BLENDUP:
		music_player->SetBlend(music_player->GetBlend() + 25);
		lcd->SetLed(LedControl::EQ0, music_player->GetBlend() / 50);
		break;
	case Input::BLENDDOWN:
		music_player->SetBlend(music_player->GetBlend() - 25);
		lcd->SetLed(LedControl::EQ0, music_player->GetBlend() / 50);
	   	break;
	case Input::SCREENSHOT:
		g_screen.ScreenShot("screenshot.bmp");
		//play_list->Save("playlist");
		break;
	}

	return true;
}


void PlayerApp::InitSettings(char *fname)
{
	//char *led;

	fprintf(stderr, "Init config\n");
	fflush(stderr);

	config.Init(".settings");
	config["REPEAT"] = "0";
	config["SHUFFLE"] = "0";
	config["CPUSCALE"] = "0";
	config["VOLUME"] = "8";
	config.Load();
	repeat = atoi(config["REPEAT"]);
	shuffle = atoi(config["SHUFFLE"]);
	cpu_scaling = atoi(config["CPUSCALE"]);
	config.Save();

	if(fname)
		screen_cfg.Init(fname);
	else
		screen_cfg.Init("default.cfg");

	fprintf(stderr, "Settings defaults\n");
	fflush(stderr);

	screen_cfg["COL0"] = "0xB2DBF8";
	screen_cfg["COL1"]= "0x5084A4";
	screen_cfg["COL2"] = "0x1F596C";

	screen_cfg["TEXTCOL"] = "0xFFFFFF";
	screen_cfg["BGCOL"] = "0x000000";
	screen_cfg["FLAGCOL"] = "0x00FF00";
	screen_cfg["MARKCOL"] = "0x000080";
	screen_cfg["DIRCOL"] = "0xFFFF80";
	screen_cfg["SONGCOL0"] = "0x438AB3";
	screen_cfg["SONGCOL1"] = "0x63AAD3";
	screen_cfg["BACKGROUND"] = "";

	screen_cfg["HMARGIN"] = "8";
	screen_cfg["VMARGIN"] = "8";
	screen_cfg["SPACING"] = "8";

//	settings.bgcol = 0x000000;
//	settings.textcol = 0xFFFFFF;
	settings.width = 320;
	settings.height = 240;


#ifdef GP2X

	screen_cfg["FIXEDFONT"] = "";
	screen_cfg["PROPFONT"] = "";
	screen_cfg["LEDBMP"] = "led.bmp";

#else

	screen_cfg["FIXEDFONT"] = "Fixed.font";
	screen_cfg["PROPFONT"] = "Proportional.font";
	screen_cfg["WIDTH"] = "640";
	screen_cfg["HEIGHT"] = "570";
	screen_cfg["LEDBMP"] = "bigled.bmp";

#endif

	fprintf(stderr, "Loading\n");
	fflush(stderr);
	screen_cfg.Load();

	for(int i=0; i<Screen::MAX_COL; i++)
	{
		if(screen_cfg[Screen::ColorNames[i]])
			settings.colors[i] = strtol(screen_cfg[Screen::ColorNames[i]], NULL, 16);
	}

	//settings.bevellight = strtol(screen_cfg["COL0"], NULL, 16);
	//settings.bevelmid = strtol(screen_cfg["COL1"], NULL, 16);
	//settings.beveldark = strtol(screen_cfg["COL2"], NULL, 16);
	settings.font0 = screen_cfg["FIXEDFONT"];
	settings.font1 = screen_cfg["PROPFONT"];
	settings.background = screen_cfg["BACKGROUND"];
	hmargin = strtol(screen_cfg["HMARGIN"], NULL, 10);
	vmargin = strtol(screen_cfg["VMARGIN"], NULL, 10);
	sep = strtol(screen_cfg["SPACING"], NULL, 10);
#ifndef GP2X
	settings.width = strtol(screen_cfg["WIDTH"], NULL, 10);
	settings.height =  strtol(screen_cfg["HEIGHT"], NULL, 10);
#endif
	//settings.led = screen_cfg["LEDBMP"];
/*
	settings.textcol = strtol(screen_cfg["TEXTCOL"], NULL, 16);
	settings.bgcol = strtol(screen_cfg["BGCOL"], NULL, 16);
	settings.flagcol = strtol(screen_cfg["FLAGCOL"], NULL, 16);
	settings.markcol = strtol(screen_cfg["MARKCOL"], NULL, 16);
	settings.dircol = strtol(screen_cfg["DIRCOL"], NULL, 16);
	settings.songcol0 = strtol(screen_cfg["SONGCOL0"], NULL, 16);
	settings.songcol1 = strtol(screen_cfg["SONGCOL1"], NULL, 16);
*/

	fprintf(stderr, "Starting\n");
	fflush(stderr);
	g_screen.Init(&settings);
	printf("OK\n");

	printf("OK\n");

}


void PlayerApp::RenderEq()
{
	short buff[2048*2];
	short buff2[1024*2];

	int l = music_player->GetNextSamples(buff, sizeof(buff)) / 2;

	int avg0 = 0;
	int avg1 = 0;

	if(l)
	{
		int i;
		for(i=0; i<l; i+=2)
		{
			avg0 += ABS(buff[i]);
			avg1 += ABS(buff[i+1]);
		}
		//avg0 /= l;
		//avg1 /= l;
		//cprintf("%d %d\n", l, avg0*2/l);
		lcd->SetLed(LedControl::BAR0, avg0 / (l*256));
		lcd->SetLed(LedControl::BAR1, avg1 / (l*256));

		for(i=0; i<l/4; i++)
			buff2[i] = (buff[i*2] + buff[i*2+1]) / 2;

		if(eq)
			eq->WriteSamples(buff2, l/4);
	}
}

bool PlayerApp::PlaySong(IFileList *list, int index)
{
	int fsize = list->GetSize(index);
	const char *name = list->GetName(index);
	IFileRef *fr = list->GetFileRef(index);

	bool ok = false;
	if(fr->GetFileName())
		ok = music_player->PlaySong(fr->GetFileName());
	else
	{
		char *tmp = (char *)malloc(fsize);
		fr->Read(tmp, 0, fsize);
		ok = music_player->PlaySong(name, tmp, fsize);
		free(tmp);
	}

	if(ok)
	{
		lcd->SetLed(LedControl::MAXSONG, music_player->GetMaxTunes());
		lcd->SetLed(LedControl::SONGNO, music_player->GetTune()+1);
		songInfo->SetName(music_player->GetName());
		songInfo->SetArtist(music_player->GetArtist());
		songInfo->SetCopyright(music_player->GetCopyright());
		songInfo->SetLength(music_player->GetLength());
		songInfo->SetFormat(music_player->GetFormat());
	}
	else
	{
		cprintf("Could not play file\n");
	}

	delete fr;

	playing = ok;

	return ok;
}

