/*  Copyright 2006 Jonas Minnberg

    This file is part of OldPlay - a portable, multiformat musicplayer.

    OldPlay is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    OldPlay is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OldPlay; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <stdarg.h>

#ifdef _WIN32
#include <windows.h>
static void dbprintf(const char *fmt, ...)
{
	char tmp[128];
	va_list vl;
	va_start(vl, fmt);
	vsprintf(tmp, fmt, vl);
	OutputDebugString(tmp);
	va_end(vl);
}
#else
static void dbprintf(const char *fmt, ...)
{
}
#endif

#include "mad.h"
#include "id3tag.h"

#include "SDL_thread.h"

#include "../plugin.h"
#include "../Fifo.h"

Fifo *mp3Fifo;
Fifo *fifo2;

static SDL_sem *read_sem;
static SDL_mutex *mutex = NULL;

static struct sound_plugin plugin;

#define IN_SIZE 4096
static unsigned char inbuf[IN_SIZE*6 + 2200];

static int playing = 0;
static FILE *mp3_file;


void mad_stream_buffer_clear(struct mad_stream *stream)
{
	int left = stream->bufend - stream->this_frame;

	int size = stream->this_frame - stream->buffer;

	memcpy((void *)stream->buffer, (void *)stream->this_frame, left);

	stream->bufend = stream->buffer + left;

	stream->this_frame -= size;
	stream->next_frame -= size;
	stream->ptr.byte -= size;
}

volatile int asked = 0;
void handle_input(struct mad_stream *stream)
{
	//dbprintf("T: Resetting buffer\n");

	SDL_mutexP(mutex);
	mad_stream_buffer_clear(stream);
	SDL_mutexV(mutex);

	//dbprintf("T: Reading\n");
	int len = fread((void *)stream->bufend, 1, IN_SIZE, mp3_file);
	//dbprintf("T: Read done\n");

	//dbprintf("T: Read %d bytes\n", len);

	if(len < IN_SIZE)
		playing = 1;

	SDL_mutexP(mutex);
	asked--;
	stream->bufend += len;
	SDL_mutexV(mutex);
}

static bool quit_thread;
int read_thread(void *data)
{
	mad_stream *stream = (mad_stream *)data;
	while(SDL_SemWait(read_sem) == 0)
	{
		//dbprintf("T: Got sema\n");
		if(quit_thread)
		{
			return 0;
		}
		handle_input(stream);
	}
	return -1;
}


void handle_output( struct mad_pcm *pcm)
{
	signed int sample;
	int nChannels = pcm->channels;
	int nSamples  = pcm->length;
	mad_fixed_t *left_ch   = pcm->samples[0];
	mad_fixed_t *right_ch  = pcm->samples[1];

	short *outp = (short *)mp3Fifo->GetDest();

	//fprintf(stderr, "Output, got %d samples\n", nSamples);

	while(nSamples--)
	{
		sample = *left_ch++;
		sample += (1L << (MAD_F_FRACBITS - 16));
		if (sample >= MAD_F_ONE)
			sample = MAD_F_ONE - 1;
		else if (sample < -MAD_F_ONE)
			sample = -MAD_F_ONE;
	   *outp++ = sample >> (MAD_F_FRACBITS + 1 - 16); 
	 
		sample = *right_ch++;
		sample += (1L << (MAD_F_FRACBITS - 16));
		if (sample >= MAD_F_ONE)
			sample = MAD_F_ONE - 1;
		else if (sample < -MAD_F_ONE)
			sample = -MAD_F_ONE;
	   *outp++ = sample >> (MAD_F_FRACBITS + 1 - 16);
	}

	mp3Fifo->Write(NULL, pcm->length*4);
}


struct mad_stream stream;
struct mad_frame frame;
struct mad_synth synth;
int sound_frames;
int total_bitrate;

static int fill_buffer(signed short *dest, int len)
{
	if(playing)
	{
		while(mp3Fifo->LeftToRead() < len)
		{
			SDL_mutexP(mutex);
			if(playing == 2 && !asked && ((stream.bufend - stream.next_frame) < IN_SIZE*4))
			{
				//SDL_mutexV(mutex);
				//dbprintf("Asking to read\n");
				asked++;
				SDL_SemPost(read_sem);
				//handle_input(&stream);
				//SDL_mutexP(mutex);
			}

			if(mad_frame_decode(&frame, &stream) == -1)
			{
				if(playing == 1)
				{
					SDL_mutexV(mutex);
					return 0;
				}
			}
			else
			{
				mad_synth_frame(&synth, &frame);
				handle_output(&synth.pcm);
			}
			SDL_mutexV(mutex);
		}
		mp3Fifo->Read(dest, len);
	}
	else
	{
		memset(dest, 0, len);
	}
	return len;
}

static SDL_Thread *thread;

static int close()
{
	int rc;
	//SDL_KillThread(thread);
	quit_thread = true;
	SDL_SemPost(read_sem);
	SDL_WaitThread(thread, &rc);

	while(SDL_SemValue(read_sem))
		SDL_SemWait(read_sem);

	mad_synth_finish(&synth);
	mad_frame_finish(&frame);
	mad_stream_finish(&stream);
	mp3Fifo->Reset();

	return 0;
}

char *get_string(struct id3_tag *id3tag, const char *what)
{
	struct id3_frame *frame = id3_tag_findframe(id3tag, what, 0);
	if(frame)
	{
		id3_field *field = &frame->fields[1];
		int nstrings = id3_field_getnstrings(field);
		for (int j=0; j<nstrings; ++j)
		{
			const id3_ucs4_t *ucs4 = id3_field_getstrings(field, j);
			id3_latin1_t *latin1 = id3_ucs4_latin1duplicate(ucs4);
			if(latin1)
				return (char*)latin1;
            //int len = strlen((char *)latin1);
		}
	}
	return NULL;
}

static int init_file(char *fname)
{
	plugin.name = "";
	plugin.artist = "";
	plugin.copyright = "";
	plugin.subtunes = 0;
	plugin.tune = 0;
	plugin.length = -2;

	fprintf(stderr, "Decoder Init\n");

    mad_stream_init(&stream);
    mad_frame_init(&frame);
    mad_synth_init(&synth);

	id3_file *id3file = id3_file_open(fname, ID3_FILE_MODE_READONLY);
	struct id3_tag *id3tag = id3_file_tag(id3file);

	char *s;
	if(s = get_string(id3tag, ID3_FRAME_TITLE))
		plugin.name = strdup(s);
	if(s = get_string(id3tag, ID3_FRAME_ARTIST))
		plugin.artist = strdup(s);
	//if(s = get_string(id3tag, ID3_FRAME_ARTIST))
	//	plugin.artist = strdup(s);


	id3_file_close(id3file);

    mp3_file = fopen(fname, "rb");
	fseek(mp3_file, 0, SEEK_END);
	int fsize = ftell(mp3_file);
	fseek(mp3_file, 0, SEEK_SET);
    int len = fread(inbuf, 1, IN_SIZE*6, mp3_file);
    mad_stream_buffer(&stream, (const unsigned char *)inbuf, len);	 

	int bitrate = -1;

	while((stream.bufend - stream.next_frame) > IN_SIZE)
	{
		if(frame.header.bitrate)
		{
			if(bitrate == -1 || bitrate == frame.header.bitrate)
				bitrate = frame.header.bitrate;
			else
				bitrate = 0;
		}

		if(mad_frame_decode(&frame, &stream) == -1)
		{
		}
		else
		{
			mad_synth_frame(&synth, &frame);
			handle_output(&synth.pcm);
		}
	}

	if(bitrate > 0)
		plugin.length = (fsize * 8 / bitrate)  * 1000;

	playing = 2;
	quit_thread = false;
	thread = SDL_CreateThread(read_thread, &stream);


	return 0;
}


static int set_position(int seconds, int subtune)
{
	return 0;
}

static int is_ext(const char *s, const char *ext)
{
	const char *sext = strrchr(s, '.');
	if(!sext)
		sext = "";
	return (strcmp(ext, sext) == 0);
}

static int can_handle(const char *name)
{
	return (is_ext(name, ".mp3"));
}

extern "C" {

#ifndef INIT_SOUND_PLUGIN
#define INIT_SOUND_PLUGIN mad_init_sound_plugin
#endif

struct sound_plugin *INIT_SOUND_PLUGIN()
{
	memset(&plugin, 0, sizeof(plugin));

	plugin.plugname = "mad";
	plugin.format = "MP3";
	plugin.freq = 44100;
	plugin.channels = 2;
	plugin.init_file = init_file;
	plugin.close = close;
	//plugin.init_data = init_data;
	plugin.fill_buffer = fill_buffer;
	plugin.set_position = set_position;
	plugin.can_handle = can_handle;
	plugin.clockfreq = 100;

	mp3Fifo = new Fifo(1024*1024);

	read_sem = SDL_CreateSemaphore(0);
	mutex = SDL_CreateMutex();

	return &plugin;
}


}

