/*  Copyright 2006 Jonas Minnberg

    This file is part of OldPlay - a portable, multiformat musicplayer.

    OldPlay is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    OldPlay is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OldPlay; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    */
#include <SDL.h>
#include <vector>
#include "FileList.h"
#include "ZipFileList.h"
#include "PlayList.h"
#include "SDLScreen.h"
#include "MusicPlayer.h"
#include "Input.h"
#include "Config.h"
#include "M3uPlayList.h"
#include "PlayerApp.h"
#include "util.h"
#include "State.h"
#include "zip/zip.h"

#ifndef _WIN32
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif

extern PlayerApp *app;

BrowserState g_browserState;

static int parent_selected = 0;
static IFileList *parent_file_list = NULL;
static ZipFileList *zip_list;
static IFileList *file_list;
static int add_index = -1;

BrowserState::BrowserState()
{
    zip_list = NULL;
}


void BrowserState::Enter()
{
    if(zip_list)
        app->pane->SetList(zip_list);
    else
        app->pane->SetList(app->file_list);

    app->pane->SetTitle(app->file_list->GetCurDir());
    app->pane->SetTitleColor(app->settings.colors[Screen::COL_BROWSER_TITLE0], app->settings.colors[Screen::COL_BROWSER_TITLE1]);
}

bool BrowserState::HandleKey(int key)
{
    ListPane *pane = app->pane;
    int selected = -1;
    IFileList *current_list = pane->GetList();

    if(key)
    {
        switch(key)
        {
            case Input::PGUP:
                pane->DoCommand(ListPane::PGUP);
                break;
            case Input::PGDOWN:
                pane->DoCommand(ListPane::PGDN);
                break;
            case Input::UP:
                pane->DoCommand(ListPane::UP);
                break;
            case Input::DOWN:
                pane->DoCommand(ListPane::DOWN);
                break;
            case Input::ENTER:
                selected = current_list->Marked();
                break;
            case Input::BACK:
                selected = -2;
                break;
            case Input::PLISTADD:
                if(add_index != current_list->Marked())
                {
                    add_index = current_list->Marked();
                    const char *name = current_list->GetName(add_index);
                    const char *path = (const char*)(current_list->GetPath(add_index));
                    if(app->music_player->CanHandle(name))
                        app->play_list->AddFile(name, current_list->GetSize(add_index), path, current_list->GetIndex(add_index));
                    else if (is_ext(name,".m3u") || is_ext(name,".pls") || is_ext(name,".opl"))
                    {
                        printf("BrowserState::HandleKey - Found m3u, appending playlist\n");
                        this->AppendPlayList(path);
                        selected = -1;
                    }
                    pane->DoCommand(ListPane::DOWN);
                }
                break;
        }
    }

    if(selected == -1)
        return true;

    add_index = -1;

    if(selected < 0)
    {
        int rc = current_list->Enter(-1);
        if(rc < 0)
        {
            delete current_list;
            current_list = parent_file_list;
            rc = parent_selected;
            parent_selected = 0;
            parent_file_list = NULL;
            zip_list = NULL;
        }
        pane->SetList(current_list);
        pane->SetTitle(current_list->GetCurDir());

        if(rc >= 0)
            current_list->SetMarked(rc);
        pane->DoCommand(ListPane::NOTHING);
    }
    else
    {
        int fsize = current_list->GetSize(selected);
        const char *name = current_list->GetName(selected);

        if(fsize == -1)
        {
            current_list->Enter(selected);
            pane->SetList(current_list);
            pane->SetTitle(current_list->GetCurDir());
            current_list->SetMarked(0);
            current_list->SetStart(0);
            pane->DoCommand(ListPane::NOTHING);
            selected = -1;
        }
        else if(is_ext(name, ".zip"))
        {
            parent_file_list = current_list;
            parent_selected = selected;
            zip_list = new ZipFileList(current_list->GetPath(selected),&(plugin_filter_zip_cb) ,app);
            current_list = zip_list;
            pane->SetList(current_list);
            pane->SetTitle(app->file_list->GetCurDir());
            selected = -1;
        }
        else if(is_ext(name, ".m3u") || is_ext(name, ".pls") || is_ext(name, ".opl"))
        {
            printf("BrowserState::HandleKey - Found m3u, clearing playlist and creating new\n");
            app->play_list->Clear();
            this->AppendPlayList(current_list->GetPath(selected));
            pane->DoCommand(ListPane::DOWN);
            selected = -1;
        }
        else
        {
            app->PlaySong(current_list, selected);
            pane->DoCommand(ListPane::DOWN);
        }
    }
    return true;
}

void BrowserState::AppendPlayList(const char* fname)
{
    M3uPlayList *m3u_list = new M3uPlayList(fname);
    bool opl = is_ext(fname,".opl");
    for (; m3u_list->Size(); m3u_list->Next())
    {
        printf("BrowserState::AppendPlayList - Checking entry\n"); 
        const char* m3u_path = m3u_list->GetPath();
        const char* m3u_name = m3u_list->GetName();
        int m3u_index = opl ? m3u_list->GetIndex() : -1;
        
        long fsize;
        if (is_ext(m3u_path, ".zip"))
        {
            // Open through zip, how bothersome
            struct zip* zipf;
            if (zipf = zip_open(m3u_path, 0,NULL))
            {
                struct zip_stat zs;
                char name[256];
                zs.name = &name[0];
                if (zip_stat_index(zipf, m3u_index, 1, &zs))
                    continue;
                fsize = zs.size;
                zip_close(zipf);
                // TODO WHY WONT IT FIND THE FILENAME?!
                /*printf("zs.name = %s\n",zs.name);
                if (!app->music_player->CanHandle(name))
                    continue;
                */
            }
            else
                continue;
        }
        else
        {
            if (!app->music_player->CanHandle(m3u_path))
                continue;
            FILE* fp = fopen(m3u_path,"rb");
            if (!fp)
                continue;
            fseek(fp,0,SEEK_END);
            fsize = ftell(fp);
            fclose(fp);
        }

        printf("BrowserState::AppendPlayList - Entry looks alright, adding\n"); 
        app->play_list->AddFile(m3u_name, fsize, m3u_path, m3u_index );
    }
    delete m3u_list;
}
