/*  Copyright 2006 Jonas Minnberg

    This file is part of OldPlay - a portable, multiformat musicplayer.

    OldPlay is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    OldPlay is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OldPlay; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <SDL.h>
#ifdef _WIN32
#include <direct.h>
#endif

#include "Files.h"
#include "FileList.h"
#include "PlayList.h"
#include "SaveList.h"
#include "SDLScreen.h"
#include "MusicPlayer.h"
#include "Equalizer.h"
#include "PlayerApp.h"
#include "Input.h"
#include "util.h"

#define ABS(x) ((x) < 0 ? (-x) : (x))

PlayerApp::~PlayerApp()
{
	char dst[8];
	config["REPEAT"]  = repeat  ? (char *)"1" : (char *)"0";
	config["SHUFFLE"] = shuffle ? (char *)"1" : (char *)"0";
	sprintf(dst, "%d", music_player->GetVolume());
	config["VOLUME"] = dst;
	config.Save();
}

PlayerApp::PlayerApp(char *fname)
{
	InitSettings(fname);

	pane         = new ListPane(0, 0, 0, 0);
	songInfo     = new SongInfoPane(0, 0, 0, 0);
	lcd          = new LedControl(screen_cfg["LEDBMP"], 0, 0);
	eq           = new Equalizer(0, 0, lcd->Width(), lcd->Width()/3);
	music_player = new MusicPlayer();
	music_player->SetVolume(atoi(config["VOLUME"]));

	lcd->SetLed(LedControl::SONGNO, 1);
	lcd->SetLed(LedControl::MAXSONG, 1);
	lcd->SetLed(LedControl::VOLUME, music_player->GetVolume());
	lcd->SetLed(LedControl::REPEAT, repeat);
	lcd->SetLed(LedControl::SHUFFLE, shuffle);

	char curdir[256];
	getcwd(curdir, sizeof(curdir));
	file_list = new FileList(curdir, &plugin_filter_cb, this);
	save_list = new SaveList(curdir, &filter_opl_cb, this);
	play_list = new PlayList(&plugin_filter_cb, this);
	zip_list  = NULL;

	fprintf(stderr, "PlayerApp::PlayerApp - Constructed\n");
}

void PlayerApp::Layout()
{
	int fh = g_screen.GetPropFont()->height;
	int w = g_screen.Width() - hmargin*2;
	int h = g_screen.Height() - vmargin*2;

	lcd->Move(w - lcd->Width() + hmargin, fh+sep + vmargin);
	lcd->Clear(true);

	pane->Move(hmargin, fh+sep + vmargin);
	pane->Resize(w - lcd->Width() - sep, h - (fh+sep));
	pane->Clear(true);

	songInfo->Move(w - lcd->Width() + hmargin, lcd->Height() + (fh+sep) + vmargin + sep);
	songInfo->Resize(lcd->Width(), fh * 11);
	songInfo->Clear(true);
	songInfo->Render();

	int y = lcd->Height() + (fh+sep) + sep + sep + songInfo->Height()  + vmargin;
	eq->Move(w - lcd->Width() + hmargin, y);
	eq->Clear(true);

	Font *font = g_screen.GetPropFont();

	int len = font_text(font, OLDPLAYVERSION, -1, -1);
	int f = font->flags;
	font->flags |= FFLG_TRANSP;
	font_setcolor(0x000000, settings.colors[Screen::COL_BEVEL_MID]);
	font_text(font, OLDPLAYVERSION, w-len+1+hmargin, h-font->height+vmargin+1);
	font_setcolor(0xFFFFFF, settings.colors[Screen::COL_BEVEL_MID]);
	font_text(font, OLDPLAYVERSION, w-len+hmargin, h-font->height+vmargin);
	font->flags = f;
}

int last_sec = -1;
int paused = 0;
bool oldplaying = false;

void PlayerApp::Render()
{
	if(paused)
	{
		if(((paused-1) % 15) == 0)
			lcd->SetLed(LedControl::PLAYING, (paused-1) % 30 );
		paused++;
	}
	if(playing != oldplaying)
	{
		lcd->SetLed(LedControl::PLAYING, playing );	
		oldplaying = playing;
	}
	int sec = music_player->GetCSeconds() / 10;
	if(last_sec != sec)
	{
		lcd->SetLed(LedControl::LENGTH, sec);
		last_sec = sec;
	}
	pane->Render();
	songInfo->Render();
	lcd->Render();

#ifdef GP2X
	if(!music_player->GetHeavy())
#endif
	{
		if(eq)
			eq->Render();
		RenderEq();
	}

}

int info_update_counter = 1; // For delay
char new_song = 0;
bool PlayerApp::Update()
{

    if (music_player->Playing() && !info_update_counter--)
    {
        info_update_counter = 10;
        lcd->SetLed(LedControl::SONGNO, music_player->GetTune()+1);
        UpdateInfo();
    }
	if(!music_player->Playing() && play_list->Size())
	{
        play_list->Next(repeat);
        if(play_list->Current() != -1)
            PlaySong(play_list, play_list->Current() );
	}
	return true;
}

char name_len, name_pos, subname_len, subname_pos, format_len, format_pos,
     artist_len, artist_pos, copyright_len, copyright_pos;
void PlayerApp::UpdateInfo()
{
#ifdef GP2X
        if (new_song)
        {
            new_song--;
            name_len      = strlen(music_player->GetName());
            subname_len   = strlen(music_player->GetSubname());
            artist_len    = strlen(music_player->GetArtist());
            copyright_len = strlen(music_player->GetCopyright());
            format_len    = strlen(music_player->GetFormat());
            name_pos = subname_pos = artist_pos = copyright_pos = format_pos = 0;
        }
        else
        {
            name_pos      = name_len      <= 16 || name_pos      + 16 >= name_len      ? 0: name_pos      + 1;
            subname_pos   = subname_len   <= 16 || subname_pos   + 16 >= subname_len   ? 0: subname_pos   + 1;
            artist_pos    = artist_len    <= 16 || artist_pos    + 16 >= artist_len    ? 0: artist_pos    + 1;
            copyright_pos = copyright_len <= 16 || copyright_pos + 16 >= copyright_len ? 0: copyright_pos + 1;
            format_pos    = format_len    <= 16 || format_pos    + 16 >= format_len    ? 0: format_pos    + 1;
        }
#endif
        songInfo->SetName(music_player->GetName()+name_pos);
        songInfo->SetSubname(music_player->GetSubname()+subname_pos);
        songInfo->SetArtist(music_player->GetArtist()+artist_pos);
        songInfo->SetCopyright(music_player->GetCopyright()+copyright_pos);
        songInfo->SetFormat(music_player->GetFormat()+format_pos);
        songInfo->SetLength(music_player->GetLength());
}

bool PlayerApp::HandleKey(int key)
{
	switch(key)
	{
	case Input::NEXT:
        if(play_list->Size())
		{
            play_list->Next(true);
			if(play_list->Current() != -1)
				PlaySong(play_list, play_list->Current() );
		}
		break;
	case Input::PREV:
        if(play_list->Size())
		{
            play_list->Prev(true);
			if(play_list->Current() != -1)
				PlaySong(play_list, play_list->Current() );
		}
		break;
	case Input::FF:
		music_player->SetTune(music_player->GetTune() + 1);
		lcd->SetLed(LedControl::SONGNO, music_player->GetTune()+1);
        UpdateInfo();
		break;
	case Input::REW:
		music_player->SetTune(music_player->GetTune() - 1);
		lcd->SetLed(LedControl::SONGNO, music_player->GetTune()+1);
		UpdateInfo();
		break;
	case Input::VOLUP:
		music_player->SetVolume(music_player->GetVolume() + 1);
		lcd->SetLed(LedControl::VOLUME, music_player->GetVolume());
		break;
	case Input::VOLDOWN:
		music_player->SetVolume(music_player->GetVolume() - 1);
		lcd->SetLed(LedControl::VOLUME, music_player->GetVolume());
		break;
	case Input::SHUFFLE:
		shuffle = !shuffle;
		lcd->SetLed(LedControl::SHUFFLE, shuffle);
		play_list->SetShuffle((bool)shuffle);
		break;
	case Input::REPEAT:
		repeat = !repeat;
		lcd->SetLed(LedControl::REPEAT, repeat);
		break;
	case Input::PAUSE:
		paused = music_player->Pause();
		break;
	case Input::BLENDUP:
		music_player->SetBlend(music_player->GetBlend() + 25);
		lcd->SetLed(LedControl::EQ0, music_player->GetBlend() / 50);
		break;
	case Input::BLENDDOWN:
		music_player->SetBlend(music_player->GetBlend() - 25);
		lcd->SetLed(LedControl::EQ0, music_player->GetBlend() / 50);
	   	break;
	case Input::SCREENSHOT:
		g_screen.ScreenShot("screenshot.bmp");
		break;
	}
	return true;
}

void PlayerApp::InitSettings(char *fname)
{
	fprintf(stderr, "PlayerApp::InitSettings\n");

	config.Init(".settings");
	config["REPEAT"]   = "0";
	config["SHUFFLE"]  = "0";
	config["CPUSCALE"] = "0";
	config["VOLUME"]   = "8";
	config.Load();
	repeat      = atoi(config["REPEAT"]);
	shuffle     = atoi(config["SHUFFLE"]);
	cpu_scaling = atoi(config["CPUSCALE"]);
	config.Save();

	if(fname)
		screen_cfg.Init(fname);
	else
		screen_cfg.Init("default.cfg");

	fprintf(stderr, "PlayerApp::InitSettings - Setting defaults\n");

	screen_cfg["COL0"] = "0xB2DBF8";
	screen_cfg["COL1"] = "0x5084A4";
	screen_cfg["COL2"] = "0x1F596C";

	screen_cfg["TEXTCOL"]  = "0xFFFFFF";
	screen_cfg["BGCOL"]    = "0x000000";
	screen_cfg["FLAGCOL"]  = "0x00FF00";
	screen_cfg["MARKCOL"]  = "0x000080";
	screen_cfg["DIRCOL"]   = "0xFFFF80";
	screen_cfg["SONGCOL0"] = "0x438AB3";
	screen_cfg["SONGCOL1"] = "0x63AAD3";
	screen_cfg["BACKGROUND"] = "";

	screen_cfg["HMARGIN"] = "8";
	screen_cfg["VMARGIN"] = "8";
	screen_cfg["SPACING"] = "8";

	settings.width  = 320;
	settings.height = 240;

#ifdef GP2X
	screen_cfg["FIXEDFONT"] = "";
	screen_cfg["PROPFONT"] = "";
	screen_cfg["LEDBMP"] = "theme/led.bmp";
#else
	screen_cfg["FIXEDFONT"] = "theme/Fixed.font";
	screen_cfg["PROPFONT"] = "theme/Proportional.font";
	screen_cfg["WIDTH"] = "640";
	screen_cfg["HEIGHT"] = "570";
	screen_cfg["LEDBMP"] = "theme/bigled.bmp";
#endif

	fprintf(stderr, "PlayerApp::InitSettings - Loading\n");
	screen_cfg.Load();

	for(int i=0; i<Screen::MAX_COL; i++)
		if(screen_cfg[Screen::ColorNames[i]])
			settings.colors[i] = strtol(screen_cfg[Screen::ColorNames[i]], NULL, 16);

	settings.font0 = screen_cfg["FIXEDFONT"];
	settings.font1 = screen_cfg["PROPFONT"];
	settings.background = screen_cfg["BACKGROUND"];
	hmargin = strtol(screen_cfg["HMARGIN"], NULL, 10);
	vmargin = strtol(screen_cfg["VMARGIN"], NULL, 10);
	sep = strtol(screen_cfg["SPACING"], NULL, 10);

#ifndef GP2X
	settings.width  = strtol(screen_cfg["WIDTH"],  NULL, 10);
	settings.height = strtol(screen_cfg["HEIGHT"], NULL, 10);
#endif

	fprintf(stderr, "PlayerApp::InitSettings - Applying settings\n");
	g_screen.Init(&settings);
	fprintf(stderr, "PlayerApp::InitSettings - OK\n");
}

void PlayerApp::RenderEq()
{
	short buff[2048*2];
	short buff2[1024*2];

	int l = music_player->GetNextSamples(buff, sizeof(buff)) / 2;

	int avg0 = 0;
	int avg1 = 0;

	if(l)
	{
		int i;
		for(i=0; i<l; i+=2)
		{
			avg0 += ABS(buff[i]);
			avg1 += ABS(buff[i+1]);
		}
		lcd->SetLed(LedControl::BAR0, avg0 / (l*256));
		lcd->SetLed(LedControl::BAR1, avg1 / (l*256));

		for(i=0; i<l/4; i++)
			buff2[i] = (buff[i*2] + buff[i*2+1]) / 2;

		if(eq)
			eq->WriteSamples(buff2, l/4);
	}
}

bool PlayerApp::PlaySong(IFileList *list, int index)
{
	bool ok = false;
	const char *name = list->GetName(index);
	const char *path = list->GetPath(index);
    int fsize = list->GetSize(index);
    int zipindex = list->GetIndex(index);

    if (zipindex >= 0) // Zip file!
    {
        fprintf(stderr, "PlayerApp::PlaySong - Loading (zipped) %s (%d)\n",path,zipindex);
        char *tmp = (char *)malloc(fsize);
        read_zip(tmp, path, fsize, zipindex);
        ok = music_player->PlaySong(name, tmp, fsize);
    }
	else if(list->GetPath(index))
    {
        fprintf(stderr, "PlayerApp::PlaySong - Loading (through init_file) %s\n",path);
		ok = music_player->PlaySong(path);
    }
	else
	{
        fprintf(stderr, "PlayerApp::PlaySong - Loading (through init_data) %s\n",path);
		char *tmp = (char *)malloc(fsize);
        read_file(tmp,list->GetPath(index),fsize);
		ok = music_player->PlaySong(name, tmp, fsize);
		free(tmp);
	}
	playing = ok;
	if(playing)
	{
		lcd->SetLed(LedControl::MAXSONG, music_player->GetMaxTunes());
		lcd->SetLed(LedControl::SONGNO, music_player->GetTune()+1);
		UpdateInfo();
	}
	else
	{
		fprintf(stderr, "PlayerApp::PlaySong - Could not play file\n");
		lcd->SetLed(LedControl::MAXSONG, 0);
		lcd->SetLed(LedControl::SONGNO, 0);
        songInfo->SetName("");
        songInfo->SetSubname("");
        songInfo->SetArtist("");
        songInfo->SetCopyright("");
        songInfo->SetFormat("Unplayable file");
        songInfo->SetLength(0);
	}
    music_player->SetSongEnd(true);
	return ok;
}

